/* Auspice functions.
 *
 * Auspice Services is copyright (c) 2000-2001 In Mean.
 *     E-mail: <auspice@auspice.org>
 *
 * This program is free but copyrighted software; see the file LICENSE for
 * details.
 */

#include "../inc/services.h"
#include "../inc/timeout.h"

static BotInfo *makebot(const char *nick);
static BotInfo *botlists[256];
static void do_bot(const char *source);
static void alpha_insert_bot(BotInfo *bi);
static void change_bot_nick(BotInfo *bi, char *newnick);
static void insert_bot(BotInfo *bi);

/**********************************************************************/
void load_bs2_db(void)
{
    FILE *f = fopen(AUTH_DB, "r");
    int i, ver;
    BotInfo *bi;

    if (!(f = open_db(s_BotServ, BOTSERV2_DB, "r", 1)))
        return;

    switch (ver = get_file_version(f, BOTSERV2_DB)) {
	case 1:
        for (i = 33; i < 256; ++i) {
            while (fgetc(f) == 1) {
                   bi = smalloc(sizeof(BotInfo));
                   if (1 != fread(bi, sizeof(BotInfo), 1, f))
                       fatal_perror("Read error on %s", BOTSERV2_DB);

                if (bi->nick)
                    bi->nick = read_string(f, BOTSERV2_DB);
                alpha_insert_bot(bi);

                if (bi->user)
                    bi->user = read_string(f, BOTSERV2_DB);
                if (bi->host)
                    bi->host = read_string(f, BOTSERV2_DB);
                if (bi->real)
                    bi->real = read_string(f, BOTSERV2_DB);
                if (bi->master)
                    bi->master = read_string(f, BOTSERV2_DB);
		if (bi->flags == 0)
		    bi->flags |= BI_USER;
            } /* while (fgetc(f) == 1) */
        } /* for (i) */
	break;
      default:
        fatal("Unsupported version number (%d) on %s", i, BOTSERV2_DB);
    } /* switch (version) */

    close_db(f, BOTSERV2_DB);
}
/**********************************************************************/
static void alpha_insert_bot(BotInfo *bi)
{
    BotInfo *bi2, *bi3;
    char *nick = bi->nick;

    for (bi3 = NULL, bi2 = botlists[tolower(*nick)];
                        bi2 && stricmp(bi2->nick, nick) < 0;
                        bi3 = bi2, bi2 = bi2->next)
        ;
    bi->prev = bi3;
    bi->next = bi2;
    if (!bi3)
        botlists[tolower(*nick)] = bi;
    else
        bi3->next = bi;
    if (bi2)
        bi2->prev = bi;
}
/**********************************************************************/
void save_bs2_db(void)
{
    FILE *f;
    int i;
    BotInfo *bi;

    if (!(f = open_db(s_BotServ, BOTSERV2_DB, "w", BS_VERSION)))
        return;
    for (i = 33; i < 256; ++i) {
        for (bi = botlists[i]; bi; bi = bi->next) {
            fputc(1, f);

            if (1 != fwrite(bi, sizeof(BotInfo), 1, f))
            fatal_perror("Write error on %s", BOTSERV2_DB);
            if (bi->nick) {
                write_string(bi->nick, f, BOTSERV2_DB);
            }
            if (bi->user) {
                write_string(bi->user, f, BOTSERV2_DB);
            }
            if (bi->host) {
                write_string(bi->host, f, BOTSERV2_DB);
            }
            if (bi->real) {
                write_string(bi->real, f, BOTSERV2_DB);
            }
            if (bi->master) {
                write_string(bi->master, f, BOTSERV2_DB);
            }
        }
        fputc(0, f);
    }
    close_db(f, BOTSERV2_DB);
}
/**********************************************************************/
static BotInfo *makebot(const char *nick)
{
    BotInfo *bi;

    bi = scalloc(sizeof(BotInfo), 1);
    bi->nick = sstrdup(nick);
    alpha_insert_bot(bi);
    return bi;
}
/**********************************************************************/
static int delbot(BotInfo *bi)
{
    do_expireabot(bi->nick);
    if (bi->next)
        bi->next->prev = bi->prev;
    if (bi->prev)
        bi->prev->next = bi->next;
    else
        botlists[tolower(*bi->nick)] = bi->next;
    if (bi->nick)
      free(bi->nick);
    if (bi->user)
      free(bi->user);
    if (bi->host)
      free(bi->host);
    if (bi->real)
      free(bi->real);
    if (bi->master)
      free(bi->master);

    free(bi);
    return 1;
}
/**********************************************************************/
BotInfo *findbot(const char *nick)
{
    BotInfo *bi;
    if (!nick || !*nick) return NULL;
    for (bi = botlists[tolower(*nick)]; bi; bi = bi->next) {
        if (stricmp(bi->nick, nick) == 0)
            return bi;
    }
    return NULL;
}
/**********************************************************************/
void sbotserv(const char *source, char *buf)
{
    BotInfo *bi;
    char *cmd, *s, *chan, *bchan;
    cmd = strtok(buf, " ");

    if (cmd) {
            if (stricmp (cmd, "\1PING") == 0)
            {
                if (!(s = strtok (NULL, "")))
                    s = "\1";
                notice(s_BotServ, source, "\1PING %s", s);

            } else if (stricmp (cmd, "\1VERSION\1") == 0) {
                if (!(s = strtok (NULL, "")))
                    s = "\1";
                notice(s_BotServ, source, "%sVERSION Auspice IRC Services %s", s, s);

            } else if (stricmp(cmd, "MUTE") == 0) {
                int i;
		if (!is_services_coder(source))
			return;
		for (i = 33; i < 256; ++i)
		for (bi = botlists[i]; bi; bi = bi->next)
		send_cmd(s_NickServ, "SVSMODE %s +d 0", bi->nick);
            } else if (stricmp(cmd, "UNMUTE") == 0) {
                int i;
		if (!is_services_coder(source))
			return;
		for (i = 33; i < 256; ++i)
		for (bi = botlists[i]; bi; bi = bi->next)
		send_cmd(s_NickServ, "SVSMODE %s -d 0", bi->nick);
            } else if (stricmp(cmd, "LIST") == 0) {
                int i;
                notice(s_BotServ, source, "\2Bot Listing\2.");
		for (i = 33; i < 256; ++i)
			for (bi = botlists[i]; bi; bi = bi->next)
			if (bi->flags & BI_USER) {
				notice(s_BotServ, source, "%-10s \2%-15s\2 (%s@%s)", "Public",bi->nick, bi->user, bi->host);
			} else if (is_oper(source)) {
				notice(s_BotServ, source, "%-10s \2%-15s\2 (%s@%s)", bi->flags & BI_MASTER?"Private":"NotSet",bi->nick, bi->user, bi->host);
			}
                notice(s_BotServ, source, "\2End OF List\2.");

            } else if (stricmp(cmd, "BOT") == 0) {
		do_bot(source);
            } else if (stricmp(cmd, "WHOIS") == 0) {
		struct tm tm;
		char timebuf[64];
                if (!is_services_admin(source)) {
                        notice(s_BotServ, source, "Permision Denied.");
                        return;
                }
                chan = strtok(NULL, " ");
                if (!chan) {
                        notice(s_BotServ, source, "Syntax: WHOIS \2Botnick\2");
                } else {
			bi = findbot(chan);
                        if (bi) {
			        char bufss[512];
				*timebuf =0;
				notice(s_BotServ, source,"Bot Nick   : %s (%s@%s) %s", bi->nick, bi->user, bi->host, bi->real);
				notice(s_BotServ, source,"Master     : %s", bi->master);
			        tm = *localtime(&bi->created);
			        strftime(timebuf, sizeof(timebuf), "%b %d %H:%M:%S %Y %Z", &tm);
			        timebuf[sizeof(timebuf)-1] = 0;
			        notice(s_BotServ, source,"Registered : %s", timebuf);
				*bufss = 0;
				if (bi->flags & BI_MASTER)
				strcat(bufss, "Private");
				if (bi->flags & BI_USER) {
			        if (*bufss)
					strcat(bufss, ", ");
				strcat(bufss, "Public");
				}
			        notice(s_BotServ, source,"Options: %s", bufss);
                        }
                }

            } else if (stricmp(cmd, "SNOOP") == 0) {
                if (!is_services_root(source)) {
                        notice(s_BotServ, source, "Permision Denied.");
                        return;
                }

                chan = strtok(NULL, " ");
                bchan = strtok(NULL, " ");

		if (!chan || !bchan) {
                        notice(s_BotServ, source, "Syntax: SNOOP \2source\2 \2dest\2");
                } else if (stricmp(chan,"NONE") == 0) {
                        aspychan = 0;
                        notice(s_BotServ, source, "Disable Channel Spy feature");
                } else {
                        if (asource)
                                free(asource);
                        asource = sstrdup(chan);
                        if (adest)
                                free(adest);
                        adest = sstrdup(bchan);
                        aspychan = 1;
                        notice(s_BotServ, source, "All message in channel %s forward to %s.", chan, bchan);
                }

            } else if (stricmp(cmd, "HELP") == 0) {
		char *subcom = strtok(NULL, " ");
		if (!subcom) {
	                notice(s_BotServ, source,"Bot Services Help Index.");
	                notice(s_BotServ, source, "- BOT     Control Bot.");
	                notice(s_BotServ, source, "- LIST    List all Services Bot.");
	                notice(s_BotServ, source, "- WHOIS   Do whois a bot.");
	                notice(s_BotServ, source, "- SNOOP   Spy channel.");
		} else if (stricmp(subcom, "BOT")==0) {
			char *subcoms = strtok(NULL, " ");
			if (!subcoms) {
				notice(s_BotServ, source,"Bot Services Help Index.");
				notice(s_BotServ, source, "- ADD      Connect a bot.");
				notice(s_BotServ, source, "- DEL      Delete Bot.");
				notice(s_BotServ, source, "- CHANGE   Change bot setting.");
				notice(s_BotServ, source, "- TRANSFER Transfer bot.");
				notice(s_BotServ, source, "- SET      Modify bot setting.");
				notice(s_BotServ, source,"Syntax: BOT <option> <arg>.");
			} else if (stricmp(subcoms, "SET")==0) {
				char *subcomss = strtok(NULL, " ");
				if (!subcomss) {
					notice(s_BotServ, source,"Bot Services Help Index.");
					notice(s_BotServ, source, "- MASTER      Allow bot as a private bot for the owner use only.");
					notice(s_BotServ, source,"Syntax: BOT SET <option> <arg>.");
				} else {
			                notice(s_BotServ, source, "No Help for this option.");
				}
			} else {
		                notice(s_BotServ, source, "No Help for this option.");
			}
		} else {
	                notice(s_BotServ, source, "No Help for this option.");
		}
            } else
                notice(s_BotServ, source, "Invalid command.");
	}
}
 
/**************************************************************/
static void do_bot(const char *source)
{
    BotInfo *bi;
    NickInfo *ni = findnick(source);
    char *cmd = strtok(NULL, " ");
    if (!cmd) {
        notice(s_BotServ, source, "Syntax: /MSG %s \2BOT\2 <command> [<parameters>].", s_BotServ);
    } else if (!is_services_admin(source) || !ni) {
        notice(s_BotServ, source, "Permision Denied.");
    } else if (!stricmp(cmd, "ADD")) {
        char *nick = strtok(NULL, " ");
        char *user = strtok(NULL, " ");
        char *host = strtok(NULL, " ");
        char *real = strtok(NULL, "");

        if (!nick || !user || !host || !real)  {
		notice(s_BotServ, source,"Syntax: BOT ADD \2botnick\2 \2username\2 \2bothost\2 \2realname\2");
		notice(s_BotServ, source, "Example: BOT ADD testbot MyNewBot botserv.auspice.org In Mean Bot");
        } else if (findbot(nick)) {
                notice(s_BotServ, source, "This bot already exist.");
	} else {
		if (finduser(nick)) {
			notice(s_BotServ, source, "This nick currently online.");
			return;
		}
                bi = makebot(nick);
                if (!bi) {
	                notice(s_BotServ, source, "Failed created bot.");
			return;
                }
                bi->user = sstrdup(user);
                bi->host = sstrdup(host);
                bi->real = sstrdup(real);
                bi->master = sstrdup(source);
                bi->created = CTime;
		bi->flags = 0;
		bi->flags |= BI_MASTER;
                bi->chancount = 0;
                BNEWNICK(bi->nick, bi->user, bi->host, bi->real, BOTMODE, 1);
                notice(s_BotServ, source, "Bot \2%s\2 connected.", bi->nick);
	}
     } else if (!stricmp(cmd, "CHANGE")) {
        char *oldnick = strtok(NULL, " ");
        char *nick = strtok(NULL, " ");
        char *user = strtok(NULL, " ");
        char *host = strtok(NULL, " ");
        char *real = strtok(NULL, "");
	int is_sr = is_services_root(source);

        if (!oldnick || !nick)
                notice(s_BotServ, source, "Syntax: BOT CHANGE \2old-bot-nick\2 \2new-bot-nick\2 [\2user\2] [\2host\2] [\2real\2]");
        else if (!(bi = findbot(oldnick)))
                notice(s_BotServ, source, "Bot %s doesn't exist", oldnick);
	else if (!is_sr && (stricmp(source, bi->master) != 0)) {
	        notice(s_BotServ, source, "Permision Denied.");
		return;
        } else {
                if (!strcmp(bi->nick, nick)
                                && ((user) ? !strcmp(bi->user, user) : 1)
                                && ((host) ? !strcmp(bi->host, host) : 1)
                                && ((real) ? !strcmp(bi->real, real) : 1))
                {
                        notice(s_BotServ, source, "Setting didn't changed.");
                        return;
                }
                if (stricmp(bi->nick, nick) && findbot(nick))
                {
                        notice(s_BotServ, source, "Bot %s already existed", nick);
                        return;
                }
                if (stricmp(bi->nick, nick) && finduser(nick))
                {
                        notice(s_BotServ, source, "nick %s currently online", nick);
                        return;
                }
                if (stricmp(bi->nick, nick))
                {
			User *u2;
                        send_cmd(NULL, "UNSQLINE %s", bi->nick);
                        if ((u2 = finduser(nick)))
                                kill_user(s_BotServ, u2->nick, "This nick is now used by Services");
                }
                if (strcmp(nick, bi->nick))
                        change_bot_nick(bi, nick);
                if (user && strcmp(user, bi->user))
                {
                        free(bi->user);
                        bi->user = sstrdup(user);
                }
                if (host && strcmp(host, bi->host))
                {
                        free(bi->host);
                        bi->host = sstrdup(host);
                }
                if (real && strcmp(real, bi->real))
                {
                        free(bi->real);
                        bi->real = sstrdup(real);
                }
                if (!user)
                        send_cmd(oldnick, "NICK %s", bi->nick);
                else
                {
			do_checkbotpart(oldnick, bi->nick);
                        send_cmd(oldnick, "QUIT :Quit: Be right back");
	                BNEWNICK(bi->nick, bi->user, bi->host, bi->real, BOTMODE, 1);
                        do_checkbot2(bi->nick);
                }
                notice(s_BotServ, source, "Bot \2%s\2 has been changed to %s!%s@%s (%s)", oldnick, bi->nick, bi->user, bi->host, bi->real);
	}
     } else if (!stricmp(cmd, "TRANSFER")) {
        char *nick = strtok(NULL, " ");
        char *master = strtok(NULL, " ");
	int is_sr = is_services_root(source);

        if (!nick || !master)
                notice(s_BotServ, source, "Syntax: BOT TRANSFER \2bot-nick\2 \2new-master-nick\2");
        else if (!(bi = findbot(nick)))
                notice(s_BotServ, source, "Bot %s doesn't exist", nick);
        else if (!findnick(master))
                notice(s_BotServ, source, "New master nick is not registered");
	else if (!is_sr && (stricmp(source, bi->master))) {
	        notice(s_BotServ, source, "Permision Denied.");
		return;
        } else {
                if (stricmp(master, bi->master))
                {
                        free(bi->master);
                        bi->master = sstrdup(master);
                }
                notice(s_BotServ, source, "Bot \2%s\2 has been changed master to (%s)", nick, bi->master);
	}
     } else if (!stricmp(cmd, "SET")) {
        char *nick = strtok(NULL, " ");
        char *opts = strtok(NULL, " ");
        char *choo = strtok(NULL, " ");
	int is_sr = is_services_root(source);

	if (!nick || !opts || !choo) {
		notice(s_BotServ, source,"Syntax: BOT SET \2botnick\2 \2option\2 \2ON|OFF\2");
		return;
	}
	bi = findbot(nick);
	if (!bi) {
		notice(s_BotServ, source,"This services bot does not exist.");
		return;
	}
	if (stricmp(opts, "MASTER")==0) {
		if (!is_sr && (stricmp(source, bi->master) != 0)) {
		        notice(s_BotServ, source, "Permision Denied.");
			return;
		}
		if (stricmp(choo, "ON")==0) {
			bi->flags |= BI_MASTER;
			bi->flags &= ~BI_USER;
			notice(s_BotServ, source,"This services bot can assigned or use by the master only.");
		} else if (stricmp(choo, "OFF")==0) {
			bi->flags |= BI_USER;
			bi->flags &= ~BI_MASTER;
			notice(s_BotServ, source,"This services bot is public for everone.");
		} else {
			notice(s_BotServ, source,"Syntax: BOT SET \2botnick\2 \2MASTER\2 \2ON|OFF\2.");
		}	
	} else {
		notice(s_BotServ, source,"Syntax: BOT SET \2botnick\2 \2option\2 \2ON|OFF\2");
	}
     } else if (!stricmp(cmd, "DEL")) {
        char *nick = strtok(NULL, " ");

        if (!nick)
		notice(s_BotServ, source,"Syntax: BOT DEL \2botnick\2");
        else if (!(bi = findbot(nick)))
                notice(s_BotServ, source, "This bot does not existed.");
        else
        {
                send_cmd(bi->nick, "QUIT :Quit: Deleted by %s!", source);
                send_cmd(NULL, "UNSQLINE %s", bi->nick);
                delbot(bi);
		notice(s_BotServ, source,"Deleted bot \2%s\2.", nick);
	}
    }
}
/**************************************************************/
void connectbot2()
{
	int i;
	BotInfo *bi;
	for (i = 33; i < 256; ++i)
		for (bi = botlists[i]; bi; bi = bi->next)
		if (bi) {
	                BNEWNICK(bi->nick, bi->user, bi->host, bi->real, BOTMODE, 1);
		}
}
/*************************************************************/
int check_ifbot(const char *abotname)
{
	if (findbot(abotname))
           return 1;
        return 0;
}
/************************************************************/
static void change_bot_nick(BotInfo *bi, char *newnick)
{
        if (bi->next)
        bi->next->prev = bi->prev;
    if (bi->prev)
        bi->prev->next = bi->next;
    else
    botlists[tolower(*bi->nick)] = bi->next;

        if (bi->nick) free(bi->nick);
        bi->nick = sstrdup(newnick);

        insert_bot(bi);
}
/***************************************/
static void insert_bot(BotInfo *bi)
{
    /* bi->prev = NULL;
    bi->next = botlist;
    if (bi->next) bi->next->prev = bi;
    botlist = bi; */

    BotInfo *ptr, *prev;

    for (prev = NULL, ptr = botlists[tolower(*bi->nick)];
                        ptr != NULL && stricmp(ptr->nick, bi->nick) < 0;
                        prev = ptr, ptr = ptr->next)
        ;
        bi->prev = prev;
    bi->next = ptr;
    if (!prev)
        botlists[tolower(*bi->nick)] = bi;
    else
        prev->next = bi;
    if (ptr)
        ptr->prev = bi;
}
/************************************************/
void botjoin2(ChannelInfo *ci)
{
        BotInfo *bi;
        Channel *c;
	if (!ci)
		return;
	c = findchan(ci->name);
	if (c && (c->havebot == 1))
		return;

        if (ci && ci->bot) {
		bi = findbot(ci->bot);
                if (bi) {
                      send_cmd(ci->bot, "JOIN %s", ci->name);
                      send_cmd(ci->bot, "MODE %s +oa %s", ci->name, ci->bot);
                      if (c)
                                c->havebot = 1;
                } else {
                        free(ci->bot);
			ci->bot = NULL;
                }
        }

}
/************************************************/
void botpart2(ChannelInfo *ci, char *newbot)
{
        Channel *c = findchan(ci->name);
        send_cmd(ci->bot, "PART %s :Be right back", ci->name);
	if (c) 
		c->havebot = 0;
	if (ci->bot) {
		free(ci->bot);
		ci->bot = sstrdup(newbot);
	}

}
/************************************************/
void bsconvert(char *nick, char *master)
{
	BotInfo *bi;
        bi = makebot(nick);
        if (!bi)
		return;
        bi->user = "BotServ";
        bi->host = sstrdup(services_host);
        bi->real = "Services Bot";
        bi->master = sstrdup(master);
        bi->created = CTime;
	bi->flags = 0;
	bi->chancount = 0;
}
/************************************************/
int is_botowner(const char *nick, const char *buf)
{
	BotInfo *bi = findbot(buf);
	if (nick && buf && bi)
		if (stricmp(bi->master, nick) == 0)
			return 1;
	return 0;
}
